package org.processmining.models.animation;

import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphEdge;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphNode;
import org.processmining.models.graphbased.directed.DirectedGraph;

public abstract class Animation {

	/*
	 * The log and graph for which this animation is created. Both fields can be
	 * removed once the deprecated methods have been removed.
	 */
	//	private XLog log;
	//	private DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph;

	/*
	 * The layout required for getting node and edge coordinates.
	 */
	private AnimationLayout layout;

	/*
	 * All animations.
	 */
	private HashMap<AbstractDirectedGraphNode, NodeAnimation> nodeAnimations;
	private HashMap<AbstractDirectedGraphEdge<?, ?>, EdgeAnimation> edgeAnimations;
	private ArrayList<NodeAnimation> nodeAnimationList;
	private ArrayList<EdgeAnimation> edgeAnimationList;
	private EdgeAnimation caseAnimation;

	/*
	 * Start and end time of the entire animation.
	 */
	private long startTime;
	private long endTime;

	/*
	 * Various animation statistics.
	 */
	private int minTraverseCount; // Minimal traverse count.
	private int maxTraverseCount; // Maximal traverse count.
	private int bestQuarterTraverseCount; // Best quarter traverse count.
	private long meanCaseDuration; // Mean case duration.
	private long fadeTime; // Fade time.
	private long meanBetweenTime; // Mean mean time between events.

	/**
	 * Create an animation object.
	 * 
	 * @deprecated
	 * @see #Animation(PluginContext, DirectedGraph, XLog)
	 * 
	 * @param graph
	 * @param log
	 */
	//	@Deprecated
	//	public Animation(
	//			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph,
	//			XLog log) {
	//		this.log = log;
	//		this.graph = graph;
	//		nodeAnimations = new HashMap<AbstractDirectedGraphNode, NodeAnimation>();
	//		edgeAnimations = new HashMap<AbstractDirectedGraphEdge<?, ?>, EdgeAnimation>();
	//		nodeAnimationList = new ArrayList<NodeAnimation>();
	//		edgeAnimationList = new ArrayList<EdgeAnimation>();
	//		startTime = Long.MAX_VALUE;
	//		endTime = Long.MIN_VALUE;
	//	}

	/**
	 * Creates an animation object.
	 * 
	 * @param context
	 * @param graph
	 * @param log
	 */
	public Animation(PluginContext context,
			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph,
			XLog log) {
		//		this.log = log;
		//		this.graph = graph;
		nodeAnimations = new HashMap<AbstractDirectedGraphNode, NodeAnimation>();
		edgeAnimations = new HashMap<AbstractDirectedGraphEdge<?, ?>, EdgeAnimation>();
		nodeAnimationList = new ArrayList<NodeAnimation>();
		edgeAnimationList = new ArrayList<EdgeAnimation>();
		startTime = Long.MAX_VALUE;
		endTime = Long.MIN_VALUE;
	}

	//	public DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> getGraph() {
	//		return graph;
	//	}

	//	public XLog getLog() {
	//		return log;
	//	}

	public AnimationLayout getLayout() {
		return layout;
	}

	public List<NodeAnimation> getNodeAnimations() {
		return nodeAnimationList;
	}

	public List<EdgeAnimation> getEdgeAnimations() {
		return edgeAnimationList;
	}

	public EdgeAnimation getCaseAnimation() {
		return caseAnimation;
	}

	public long getStart() {
		return startTime;
	}

	public long getEnd() {
		return endTime;
	}

	public int getMinTraverseCount() {
		return minTraverseCount;
	}

	public int getMaxTraverseCount() {
		return maxTraverseCount;
	}

	public int getBestQuarterTraverseCount() {
		return bestQuarterTraverseCount;
	}

	public long getFadeTime() {
		return fadeTime;
	}

	public long getMeanBetweenTime() {
		return meanBetweenTime;
	}

	public long getMeanCaseDuration() {
		return meanCaseDuration;
	}

	/**
	 * 
	 * @param left
	 *            The start time of the given interval (inclusive).
	 * @param right
	 *            The end time of the given interval (exclusive).
	 * @return The amount of activity (number of state key frames) occurring in
	 *         the given time interval.
	 */
	public int getActivityBetween(long left, long right) {
		int activity = 0;
		for (NodeAnimation anim : nodeAnimations.values()) {
			activity += anim.getActivityBetween(left, right);
		}
		return activity;
	}

	/**
	 * Builds the entire animation.
	 * 
	 * @deprecated
	 * @see #Animation(PluginContext, DirectedGraph, XLog)
	 * 
	 * @param context
	 *            Progress will be shown in the progress bar of this context.
	 * @throws IndexOutOfBoundsException
	 */
	//	@Deprecated
	//	public void initialize(PluginContext context) throws IndexOutOfBoundsException {
	//		initialize(context, graph, log);
	//	}

	/**
	 * Builds the entire animation.
	 * 
	 * @param context
	 *            Progress will be shown inht eprogress bar of this context.
	 * @param graph
	 *            The animation will be constructed for this graph.
	 * @param log
	 *            And for this log.
	 * @throws IndexOutOfBoundsException
	 */
	public void initialize(PluginContext context,
			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph,
			XLog log) throws IndexOutOfBoundsException {
		/*
		 * Get log info and progress bar.
		 */
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log);
		Progress progress = context.getProgress();

		/*
		 * Obtain the graph layout. Required for creating animations.
		 */
		progress.setValue(-1);
		progress.setCaption("Obtaining graph layout...");
		layout = new AnimationLayout(context, graph);

		/*
		 * Initialize all animations.
		 */
		progress.setCaption("Initializing animation...");
		for (AbstractDirectedGraphNode state : graph.getNodes()) {
			/*
			 * Initialize a node animation for this node.
			 */
			NodeAnimation anim = new NodeAnimation(state, layout.getCoordinates(state));
			nodeAnimations.put(state, anim);
			nodeAnimationList.add(anim);
		}
		for (AbstractDirectedGraphEdge<?, ?> edge : graph.getEdges()) {
			/*
			 * Initialize a edge animation for this edge.
			 */
			EdgeAnimation anim = new EdgeAnimation(edge, new AnimationCurve(layout.getCoordinates(edge)));
			edgeAnimations.put(edge, anim);
			edgeAnimationList.add(anim);
		}
		/*
		 * Initialize the case animation.
		 */
		this.caseAnimation = new EdgeAnimation(null, null);

		/*
		 * Set bounds for progress bar.
		 */
		progress.setMinimum(0);
		progress.setMaximum(logInfo.getNumberOfEvents());

		/*
		 * Build animations for all traces.
		 */
		for (int i = 0; i < logInfo.getNumberOfTraces(); i++) {
			progress.setCaption("Reading log: trace " + (i + 1) + " / " + logInfo.getNumberOfTraces());
			createAnimations(log.get(i), progress);
		}

		/*
		 * Finalize the entire animation.
		 */
		progress.setValue(-1);
		progress.setCaption("Preparing playback...");
		for (EdgeAnimation anim : edgeAnimations.values()) {
			anim.finalizeData();
		}
		for (NodeAnimation anim : nodeAnimations.values()) {
			anim.finalizeData();
		}
		this.caseAnimation.finalizeData();

		/*
		 * Compute animation statistics.
		 */
		int[] traverseCounts = new int[edgeAnimationList.size()];
		for (int i = 0; i < traverseCounts.length; i++) {
			traverseCounts[i] = edgeAnimationList.get(i).getTraverseCount();
		}
		if (traverseCounts.length > 0) {
			Arrays.sort(traverseCounts);
			this.minTraverseCount = traverseCounts[0];
			this.maxTraverseCount = traverseCounts[traverseCounts.length - 1];
			int bqtcIndex = (int) (traverseCounts.length * 0.75f);
			this.bestQuarterTraverseCount = traverseCounts[bqtcIndex];
		} else {
			this.minTraverseCount = 0;
			this.maxTraverseCount = 0;
			this.bestQuarterTraverseCount = 0;
		}

		meanBetweenTime = 0;
		meanCaseDuration = 0;
		long start, end;
		for (XTrace trace : log) {
			if (trace.isEmpty()) {
				continue;
			}
			// Time of first event in trace.
			start = AnimationLog.getEventTime(trace.get(0)).getTime();
			// Time of last event in trace.
			end = AnimationLog.getEventTime(trace.get(trace.size() - 1)).getTime();
			if (trace.size() >= 2) {
				// Accumulate mean time between events.
				meanBetweenTime += (end - start) / (trace.size() - 1);
				// Accumulate case duration.
				meanCaseDuration += (end - start);
			} else {
				meanBetweenTime += 1000;
				meanCaseDuration += 1000;
			}
		}
		// Mean mean time between events.
		meanCaseDuration /= logInfo.getNumberOfTraces();
		// Mean case duration.
		meanBetweenTime /= logInfo.getNumberOfTraces();
		// Set fade time to approx. 5 events.
		fadeTime = meanBetweenTime * 5;
		// Allow for some time before the animation really starts.
		if (startTime > 1000) {
			startTime -= 1000; 
		} else {
			startTime = 0;
		}
		// Allow final event to fade out.
		endTime += fadeTime;
	}

	/**
	 * Updates the boundaries for the entire animation, given that an event has
	 * occurred at the given time.
	 * 
	 * @param time
	 *            The given time.
	 */
	protected void updateBoundaries(long time) {
		/*
		 * Update start time.
		 */
		if (time < startTime) {
			startTime = time;
		}
		/*
		 * Update end time.
		 */
		if (time > endTime) {
			endTime = time;
		}
	}

	protected Date getEventTime(XEvent event) {
		return AnimationLog.getEventTime(event);
	}

	protected void addTokenAnimation(TokenAnimation anim) {
		caseAnimation.addTokenAnimation(anim);
	}

	protected void addKeyframe(AbstractDirectedGraphNode node, NodeAnimationKeyframe frame) {
		nodeAnimations.get(node).addKeyframe(frame);
	}

	protected void addTokenAnimation(AbstractDirectedGraphEdge<?, ?> edge, TokenAnimation anim) {
		edgeAnimations.get(edge).addTokenAnimation(anim);
	}

	/**
	 * Creates animations for the given trace. Shows progress using the given
	 * progress bar.
	 * 
	 * @param trace
	 *            The given trace.
	 * @param progress
	 *            The given progress bar.
	 * @throws IndexOutOfBoundsException
	 */
	protected abstract void createAnimations(XTrace trace, Progress progress) throws IndexOutOfBoundsException;

	public abstract void paintNodeBackground(AbstractDirectedGraphNode node, Graphics2D g2d, double x, double y,
			double width, double height);

	public abstract void paintNodeBorder(AbstractDirectedGraphNode node, Graphics2D g2d, double x, double y,
			double width, double height);

	public abstract void paintNodeText(AbstractDirectedGraphNode node, Graphics2D g2d, double x, double y,
			double width, double height);

	public abstract void paintTokenLabel(AbstractDirectedGraphEdge<?, ?> edge, XTrace trace, Graphics2D g2d, double x,
			double y);

	public abstract float getActivity(long modelTime, long maxTaskDelay);
}
