package org.processmining.models.animation;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import org.jgraph.util.Spline2D;

public class AnimationCurve {
	/**
	 * X and Y coordinates of the control points
	 */
	private double[] controlX;
	private double[] controlY;
	/**
	 * X and Y coordinates of the simplified / approximated points
	 */
	private double[] pointsX;
	private double[] pointsY;
	/**
	 * Relative percentage of spline magnitude for each simplified /
	 * approximated point
	 */
	private double[] pointsPercentage;
	/**
	 * Bounding box of the curve
	 */
	private Rectangle2D bounds;

	public AnimationCurve(Point2D[] coords) {
		initialize(coords);
	}

	protected void initialize(Point2D[] controlPoints) {
		int numPoints = controlPoints.length;
		controlX = new double[numPoints];
		controlY = new double[numPoints];
		for (int i = 0; i < numPoints; i++) {
			controlX[i] = controlPoints[i].getX();
			controlY[i] = controlPoints[i].getY();
		}
		double[][] coords = computeSpline(controlPoints);
		this.pointsX = coords[0];
		this.pointsY = coords[1];
		// derive percentages for control points
		analyzeCurve();
	}

	public double[] getPointsX() {
		return pointsX;
	}

	public double[] getPointsY() {
		return pointsY;
	}

	public double[] getControlPointsX() {
		return controlX;
	}

	public double[] getControlPointsY() {
		return controlY;
	}

	public double[] getRelativePoint(float percentage) {
		for (int i = 0; i < pointsPercentage.length; i++) {
			if (pointsPercentage[i] == percentage) {
				return new double[] { pointsX[i], pointsY[i] };
			} else if (pointsPercentage[i] > percentage) {
				// between i'th and i-1'th point
				double innerPercentage = (percentage - pointsPercentage[i - 1])
						/ (pointsPercentage[i] - pointsPercentage[i - 1]);
				return interpolate(pointsX[i - 1], pointsY[i - 1], pointsX[i], pointsY[i], innerPercentage);
			}
		}
		// should not occur!
		return null;
	}

	public double[][] getPartialPath(double percentage) {
		// look for first control point index having a higher percentage
		// than requested
		int greaterIndex = 0;
		for (int i = 0; i < pointsPercentage.length; i++) {
			if (pointsPercentage[i] >= percentage) {
				greaterIndex = i;
				break;
			}
		}
		// copy the pre-simplified points of the approximated path, just
		// before the requested percentage
		double[][] result = new double[2][greaterIndex + 1];
		for (int i = 0; i < greaterIndex; i++) {
			result[0][i] = pointsX[i];
			result[1][i] = pointsY[i];
		}
		// add relative point between result control points to the set
		if (pointsPercentage[greaterIndex] == percentage) {
			result[0][greaterIndex] = pointsX[greaterIndex];
			result[1][greaterIndex] = pointsY[greaterIndex];
		} else {
			double innerPercentage = (percentage - pointsPercentage[greaterIndex - 1])
					/ (pointsPercentage[greaterIndex] - pointsPercentage[greaterIndex - 1]);
			double[] interpolated = interpolate(pointsX[greaterIndex - 1], pointsY[greaterIndex - 1],
					pointsX[greaterIndex], pointsY[greaterIndex], innerPercentage);
			result[0][greaterIndex] = interpolated[0];
			result[1][greaterIndex] = interpolated[1];
		}
		return result;
	}

	public Rectangle2D boundingBox() {
		return bounds;
	}

	protected Rectangle2D createBoundingBox(double x1, double y1, double x2, double y2) {
		double tmp = 0;

		if (x1 > x2) {
			tmp = x1;
			x1 = x2;
			x2 = tmp;
		} else if (x1 == x2) {
			x2 += 1f;
		}
		if (y1 > y2) {
			tmp = y1;
			y1 = y2;
			y2 = tmp;
		} else if (y1 == y2) {
			y2 += 1f;
		}
		return new Rectangle2D.Double(x1, y1, x2 - x1, y2 - y1);
	}

	public double distance(Point2D point) {
		double distance = Double.MAX_VALUE;
		double current;
		for (int i = 1; i < pointsX.length; i++) {
			current = distance(pointsX[i - 1], pointsY[i - 1], pointsX[i], pointsY[i], (float) point.getX(),
					(float) point.getY());
			if (current < distance) {
				distance = current;
			}
		}
		return distance;
	}

	protected double distance(double x1, double y1, double x2, double y2, double pX, double pY) {
		// compute squared line segment length / magnitude
		double sqLineMag = squaredLineMagnitude(x1, y1, x2, y2);
		// compute u value
		// (as in Paul Bourke's article/s at:
		// http://local.wasp.uwa.edu.au/~pbourke/geometry/pointline/)
		double u = ((pX - x1) * (x2 - x1) + (pY - y1) * (y2 - y1)) / sqLineMag;
		double result;
		if (u <= 0 || u > 1) {
			// shortest distance is not between end points; compute
			// shortest distance to end points
			double dist1 = squaredLineMagnitude(pX, pY, x1, y1);
			double dist2 = squaredLineMagnitude(pX, pY, x2, y2);
			if (dist1 < dist2) {
				result = dist1;
			} else {
				result = dist2;
			}
		} else {
			// shortest distance is between end points
			// use the described formula
			double iX = x1 + u * (x2 - x1);
			double iY = y1 + u * (y2 - y1);
			result = squaredLineMagnitude(pX, pY, iX, iY);
		}
		// return square root of result (actual value)
		return Math.sqrt(result);
	}

	protected double squaredLineMagnitude(double x1, double y1, double x2, double y2) {
		return ((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1));
	}

	protected void analyzeCurve() {
		// percentage calculation
		float total = 0.0f;
		this.pointsPercentage = new double[pointsX.length];
		pointsPercentage[0] = 0.0f;
		for (int i = 1; i < pointsX.length; i++) {
			total += distance(pointsX[i - 1], pointsY[i - 1], pointsX[i], pointsY[i]);
			pointsPercentage[i] = total;
		}
		for (int i = 1; i < pointsX.length; i++) {
			pointsPercentage[i] /= total;
		}
		// bounding box calculation
		double x1 = Float.MAX_VALUE;
		double y1 = Float.MAX_VALUE;
		double x2 = Float.MIN_VALUE;
		double y2 = Float.MIN_VALUE;
		for (int i = 0; i < pointsX.length; i++) {
			if (pointsX[i] < x1) {
				x1 = pointsX[i];
			}
			if (pointsX[i] > x2) {
				x2 = pointsX[i];
			}
			if (pointsY[i] < y1) {
				y1 = pointsY[i];
			}
			if (pointsY[i] > y2) {
				y2 = pointsY[i];
			}
		}
		// prevent zero-area bounding box
		double bWidth = x2 - x1;
		if (bWidth <= 0) {
			bWidth = 1f;
		}
		double bHeight = y2 - y1;
		if (bHeight <= 0) {
			bHeight = 1f;
		}
		this.bounds = new Rectangle2D.Double(x1, y1, bWidth, bHeight);
	}

	protected double[][] computeSpline(Point2D[] controlPoints) {

		int n = controlPoints.length;
		//Point2D p0 = controlPoints[0];
		Point2D pe = controlPoints[n - 1];
		//Point2D p1 = controlPoints[1];
		Point2D p2 = controlPoints[n - 2];
		Spline2D spline;
		int numOfSplinePoints;
		if (n > 2) { //curve
			numOfSplinePoints = 81; // 1/t +1 ; t=0.0125
		} else {
			numOfSplinePoints = 2; // straight line
		}
		double res[][] = new double[2][numOfSplinePoints];
		//according to the JGraph
		if (n > 2) {
			spline = new Spline2D(controlPoints);
			double[] point = spline.getPoint(0.9875);
			// Extrapolate p2 away from the end point, pe, to avoid integer
			// rounding errors becoming too large when creating the line end
			double scaledX = pe.getX() - ((pe.getX() - point[0]) * 128);
			double scaledY = pe.getY() - ((pe.getY() - point[1]) * 128);
			p2.setLocation(scaledX, scaledY);
			int i = 0;
			for (double t = 0; t < 1; t += 0.0125) {
				double[] xy = spline.getPoint(t);
				res[0][i] = xy[0];
				res[1][i] = xy[1];
				i++;
				//	view.sharedPath.lineTo((float) xy[0], (float) xy[1]);
			}
		} else { //at least 2 points for an edge.  draw straightline
			for (int i = 0; i < n; i++) {
				res[0][i] = controlPoints[i].getX();
				res[1][i] = controlPoints[i].getY();
			}
		}
		return res;
	}

	protected double[] interpolate(double x1, double y1, double x2, double y2, double percentage) {
		double[] interpolated = new double[2];
		double counterPercentage = 1.0f - percentage;
		interpolated[0] = (counterPercentage * x1) + (percentage * x2);
		interpolated[1] = (counterPercentage * y1) + (percentage * y2);
		return interpolated;
	}

	protected double distance(double x1, double y1, double x2, double y2) {
		// behold the power of Pythagoras..
		double a = x1 - x2;
		double b = y1 - y2;
		return Math.sqrt((a * a) + (b * b));
	}

	public Object clone() {
		AnimationCurve clone = null;
		try {
			clone = (AnimationCurve) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
			return null;
		}
		clone.controlX = controlX.clone();
		clone.controlY = controlY.clone();
		clone.pointsX = pointsX.clone();
		clone.pointsY = pointsY.clone();
		clone.pointsPercentage = pointsPercentage.clone();
		clone.bounds = (Rectangle2D) bounds.clone();
		return clone;
	}

}
