package org.processmining.models.animation;

import java.awt.geom.Dimension2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.SwingConstants;

import org.jgraph.graph.AttributeMap.SerializablePoint2D;
import org.jgraph.graph.GraphConstants;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphEdge;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphNode;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.jgraph.ProMJGraph;
import org.processmining.models.jgraph.ProMJGraphVisualizer;
import org.processmining.models.jgraph.elements.ProMGraphCell;
import org.processmining.models.jgraph.elements.ProMGraphEdge;
import org.processmining.models.jgraph.elements.ProMGraphPort;
import org.processmining.models.jgraph.views.JGraphEdgeView;
import org.processmining.models.jgraph.views.JGraphShapeView;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;

public class AnimationLayout {

	/*
	 * Width and height for the layout.
	 */
	protected double bBoxWidth;
	protected double bBoxHeight;

	/*
	 * Node and edge coordinates. For a node, these coordinates comprises of x,
	 * y, width, and height. For an edge, these coordinates comprises of a list
	 * of points (x and y).
	 */
	protected HashMap<DirectedGraphNode, Rectangle2D> nodeCoordinates;
	protected HashMap<DirectedGraphEdge<?, ?>, Point2D[]> edgeCoordinates;

	/**
	 * Builds the layout for the given graph.
	 */
	public AnimationLayout(PluginContext context,
			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph) {
		GraphLayoutConnection layout;
		try {
			context.log("[Animation] Obtain layout for " + graph.hashCode());
			layout = context.getConnectionManager().getFirstConnection(GraphLayoutConnection.class, context, graph);
			context.log("[Animation] Process layout");
			processLayout(graph, layout);
		} catch (ConnectionCannotBeObtained e) {

			/*
			 * Get a jgraph for this graph.
			 */
			context.log("[Animation] Create layout for " + graph.hashCode());
			ProMJGraph jgraph = ProMJGraphVisualizer.instance().visualizeGraph(context.createChildContext("visualize graph"), graph).getGraph();
			/*
			 * Layout this jgraph.
			 */
			JGraphFacade facade = new JGraphFacade(jgraph);
			layOutFMJGraph(graph, jgraph, facade);
			/*
			 * Get the coordinates and statistics from this layout.
			 */
			context.log("[Animation] Process layout");
			processLayout(facade);
		}
	}

	/**
	 * 
	 * @return The width of the graph.
	 */
	public double getBoundsWidth() {
		return bBoxWidth;
	}

	/**
	 * 
	 * @return The height of the graph.
	 */
	public double getBoundsHeight() {
		return bBoxHeight;
	}

	/**
	 * 
	 * @param node
	 *            The given node.
	 * @return The rectangle of the given node.
	 */
	public Rectangle2D getCoordinates(AbstractDirectedGraphNode node) {
		return nodeCoordinates.get(node);
	}

	/**
	 * 
	 * @param edge
	 *            The given edge.
	 * @return The path of the given edge. The path consists of a series of
	 *         positions (x, y).
	 */
	public Point2D[] getCoordinates(AbstractDirectedGraphEdge<?, ?> edge) {
		return edgeCoordinates.get(edge);
	}

	/*
	 * Quite crude, but sometimes we get a graph that equals the graph we want,
	 * but is really a different object. For this reason, we allow the animator
	 * to match label names in the graphs if the direct match fails.
	 */
//	private Point2D getPositionBasedOnLabel(AbstractDirectedGraphNode node, GraphLayoutConnection layout) {
//		Point2D position = new Point2D.Double(10, 10);
//		for (Object object : layout.getGraph().getNodes()) {
//			if (object instanceof AbstractDirectedGraphNode) {
//				AbstractDirectedGraphNode graphNode = (AbstractDirectedGraphNode) object;
//				if (graphNode.getLabel().equals(node.getLabel())) {
//					return layout.getPosition(graphNode);
//				}
//			}
//		}
//		return position;
//	}
//
//	private List<Point2D> getPointsBasedOnLabel(AbstractDirectedGraphEdge<?, ?> edge, GraphLayoutConnection layout) {
//		for (Object object : layout.getGraph().getEdges()) {
//			if (object instanceof AbstractDirectedGraphEdge<?, ?>) {
//				AbstractDirectedGraphEdge<?, ?> graphEdge = (AbstractDirectedGraphEdge<?, ?>) object;
//				if (graphEdge.getLabel().equals(edge.getLabel())
//						&& graphEdge.getSource().getLabel().equals(edge.getSource().getLabel())
//						&& graphEdge.getTarget().getLabel().equals(edge.getTarget().getLabel())) {
//					return layout.getEdgePoints(graphEdge);
//				}
//			}
//		}
//		return new ArrayList<Point2D>();
//	}

	/*
	 * Get the layout information from the original graph.
	 */
	private void processLayout(
			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph,
			GraphLayoutConnection layout) {
		bBoxWidth = 0;
		bBoxHeight = 0;
		nodeCoordinates = new HashMap<DirectedGraphNode, Rectangle2D>();
		edgeCoordinates = new HashMap<DirectedGraphEdge<?, ?>, Point2D[]>();

		/*
		 * Initialize bounding box.
		 */
		double minX = Integer.MAX_VALUE;
		double minY = Integer.MAX_VALUE;
		double maxX = Integer.MIN_VALUE;
		double maxY = Integer.MIN_VALUE;

		for (AbstractDirectedGraphNode node : graph.getNodes()) {
			Point2D position = layout.getPosition(node);
//			if (position == null) {
//				position = getPositionBasedOnLabel(node, layout); //new Point2D.Double(10,10);
//			}

			Dimension2D size = layout.getSize(node);
			Rectangle2D rect = new Rectangle2D.Double();
			rect.setFrame(position, size);

			nodeCoordinates.put(node, rect);
			/*
			 * Update bounding box.
			 */
			minX = Math.min(minX, rect.getX());
			minY = Math.min(minY, rect.getY());
			maxX = Math.max(maxX, rect.getX() + rect.getWidth());
			maxY = Math.max(maxY, rect.getY() + rect.getHeight());
		}
		/*
		 * Set width and height of layout. Note that edges are ignored in this
		 * width and height.
		 */
		bBoxWidth = maxX - minX;
		bBoxHeight = maxY - minY;

		for (AbstractDirectedGraphEdge<?, ?> edge : graph.getEdges()) {
			List<Point2D> edgePoints = layout.getEdgePoints(edge);
//			if (edgePoints == null) {
//				edgePoints = getPointsBasedOnLabel(edge, layout);
//			}
			Rectangle2D sourceCoordinates = getCoordinates(edge.getSource());
			Rectangle2D targetCoordinates = getCoordinates(edge.getTarget());
			Point2D path[] = new Point2D.Double[edgePoints.size() + 2];
			path[0] = new Point2D.Double(sourceCoordinates.getCenterX(), sourceCoordinates.getCenterY());
			for (int i = 0; i < edgePoints.size(); i++) {
				path[i + 1] = new Point2D.Double(edgePoints.get(i).getX(), edgePoints.get(i).getY());
			}
			path[path.length - 1] = new Point2D.Double(targetCoordinates.getCenterX(), targetCoordinates.getCenterY());
			edgeCoordinates.put(edge, path);
		}
	}

	/**
	 * Create a layout for this JGraph. Initializes and builds the facade.
	 */
	private void layOutFMJGraph(
			DirectedGraph<? extends AbstractDirectedGraphNode, ? extends AbstractDirectedGraphEdge<?, ?>> graph,
			ProMJGraph jgraph, JGraphFacade facade) {
		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();
		layout.setDeterministic(false);
		layout.setCompactLayout(false);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(20);
		layout.setOrientation(graph.getAttributeMap().get(AttributeMap.PREF_ORIENTATION, SwingConstants.SOUTH));

		facade.setOrdered(true);
		facade.setEdgePromotion(true);
		facade.setIgnoresCellsInGroups(false);
		facade.setIgnoresHiddenCells(false);
		facade.setIgnoresUnconnectedCells(false);
		facade.setDirected(false);
		facade.resetControlPoints();

		facade.run(layout, false);

		fixParallelTransitions(facade, 15);

		Map<?, ?> nested = facade.createNestedMap(true, false);

		jgraph.getGraphLayoutCache().edit(nested);
		jgraph.setUpdateLayout(layout);
	}

	/**
	 * Fix the position of parallel transitions.
	 * 
	 * @param facade
	 * @param spacing
	 */
	private void fixParallelTransitions(JGraphFacade facade, double spacing) {
		ArrayList<Object> edges = getEdges(facade);
		for (Object edge : edges) {
			List<Object> points = getPoints(facade, edge);
			if (points.size() != 2) {
				continue;
			}
			Object sourceCell = facade.getSource(edge);
			Object targetCell = facade.getTarget(edge);
			Object sourcePort = facade.getSourcePort(edge);
			Object targetPort = facade.getTargetPort(edge);
			Object[] between = facade.getEdgesBetween(sourcePort, targetPort, false);
			if ((between.length == 1) && !(sourcePort == targetPort)) {
				continue;
			}
			Rectangle2D sCP = facade.getBounds(sourceCell);
			Rectangle2D tCP = facade.getBounds(targetCell);
			Point2D sPP = GraphConstants.getOffset(((ProMGraphPort) sourcePort).getAttributes());

			if (sPP == null) {
				sPP = new Point2D.Double(sCP.getCenterX(), sCP.getCenterY());
			}
			Point2D tPP = GraphConstants.getOffset(((ProMGraphPort) targetPort).getAttributes());
			// facade.getBounds(sourcePort);

			if (tPP == null) {
				tPP = new Point2D.Double(tCP.getCenterX(), tCP.getCenterY());
			}

			if (sourcePort == targetPort) {
				assert (sPP.equals(tPP));
				double x = sPP.getX();
				double y = sPP.getY();
				for (int i = 2; i < between.length + 2; i++) {
					List<Point2D> newPoints = new ArrayList<Point2D>(5);
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y));
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x, y - (2 * spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing), y - (spacing / 2 + i * spacing)));
					facade.setPoints(between[i - 2], newPoints);
				}

				continue;
			}

			double dx = (sPP.getX()) - (tPP.getX());
			double dy = (sPP.getY()) - (tPP.getY());
			double mx = (tPP.getX()) + dx / 2.0;
			double my = (tPP.getY()) + dy / 2.0;
			double slope = Math.sqrt(dx * dx + dy * dy);
			for (int i = 0; i < between.length; i++) {
				List<Point2D> newPoints = new ArrayList<Point2D>(3);
				double pos = 2 * i - (between.length - 1);
				if (facade.getSourcePort(between[i]) == sourcePort) {
					newPoints.add(sPP);
					newPoints.add(tPP);
				} else {
					newPoints.add(tPP);
					newPoints.add(sPP);
				}
				if (pos != 0) {
					pos = pos / 2;
					double x = mx + pos * spacing * dy / slope;
					double y = my - pos * spacing * dx / slope;
					newPoints.add(1, new SerializablePoint2D.Double(x, y));
				}
				facade.setPoints(between[i], newPoints);
			}
		}
	}

	/**
	 * Process the layout.
	 * 
	 * @param facade
	 */
	private void processLayout(JGraphFacade facade) {
		bBoxWidth = 0;
		bBoxHeight = 0;
		nodeCoordinates = new HashMap<DirectedGraphNode, Rectangle2D>();
		edgeCoordinates = new HashMap<DirectedGraphEdge<?, ?>, Point2D[]>();

		/*
		 * Initialize bounding box.
		 */
		double minX = Integer.MAX_VALUE;
		double minY = Integer.MAX_VALUE;
		double maxX = Integer.MIN_VALUE;
		double maxY = Integer.MIN_VALUE;

		/*
		 * Get and store the coordinates and dimensions for all nodes. As a side
		 * effect, the bounding box is computed.
		 */
		ArrayList<Object> vertices = getVertices(facade);
		for (Object vertex : vertices) {

			if (vertex instanceof ProMGraphCell) {
				/*
				 * Get bounds for this vertex.
				 */
				ProMGraphCell node = (ProMGraphCell) vertex;
				JGraphShapeView nodeView = node.getView();
				Rectangle2D nodeBound = nodeView.getBounds();

				/*
				 * Store position and dimension for this node.
				 */
				nodeCoordinates.put(node.getNode(), nodeBound);

				/*
				 * Update bounding box.
				 */
				minX = Math.min(minX, nodeBound.getX());
				minY = Math.min(minY, nodeBound.getY());
				maxX = Math.max(maxX, nodeBound.getX() + nodeBound.getWidth());
				maxY = Math.max(maxY, nodeBound.getY() + nodeBound.getHeight());
			}
		}
		/*
		 * Set width and height of layout. Note that edges are ignored in this
		 * width and height.
		 */
		bBoxWidth = maxX - minX;
		bBoxHeight = maxY - minY;

		/*
		 * Get and store the paths for all edges.
		 */
		ArrayList<Object> edges = getEdges(facade);
		for (Object edge : edges) {
			/*
			 * Get control points for this edge.
			 */
			ProMGraphEdge graphEdge = (ProMGraphEdge) edge;
			JGraphEdgeView edgeView = graphEdge.getView();
			int size = edgeView.getPointCount();

			/*
			 * Convert control points into path (array of points).
			 */
			Point2D[] ctrlPoints = new Point2D[size];
			for (int i = 0; i < size; i++) {
				Point2D controlPoint = edgeView.getPoint(i);
				if (controlPoint != null) {
					double x = controlPoint.getX();
					double y = controlPoint.getY();
					ctrlPoints[i] = new Point2D.Double(x, y);
				}
			}

			/*
			 * Store path for this edge.
			 */
			edgeCoordinates.put(graphEdge.getEdge(), ctrlPoints);
		}
	}

	/*
	 * The next three methods have been introduced to localize the effects of
	 * the suppresWarnings annotation.
	 */

	@SuppressWarnings("unchecked")
	private ArrayList<Object> getEdges(JGraphFacade facade) {
		return new ArrayList<Object>(facade.getEdges());
	}

	@SuppressWarnings("unchecked")
	private List<Object> getPoints(JGraphFacade facade, Object edge) {
		return facade.getPoints(edge);
	}

	@SuppressWarnings("unchecked")
	private ArrayList<Object> getVertices(JGraphFacade facade) {
		return new ArrayList<Object>(facade.getVertices());
	}

}
