package org.processmining.models.animation;

import java.util.ArrayList;
import java.util.List;

import org.processmining.models.graphbased.directed.AbstractDirectedGraphEdge;

public class EdgeAnimation {
	private final AbstractDirectedGraphEdge<?, ?> edge;
	private final AnimationCurve curve;

	private List<EdgeAnimationKeyframe> keyframeList;

	private long[] keyframeTimes;
	private EdgeAnimationKeyframe[] keyframes;

	private long bufferKeyframeTime;
	private EdgeAnimationKeyframe bufferKeyframe;

	public EdgeAnimation(AbstractDirectedGraphEdge<?, ?> edge, AnimationCurve curve) {
		this.edge = edge;
		this.curve = curve;
		keyframeList = new ArrayList<EdgeAnimationKeyframe>();
	}

	public AbstractDirectedGraphEdge<?, ?> getEdge() {
		return edge;
	}
	
	public AnimationCurve getCurve() {
		return curve;
	}

	public synchronized EdgeAnimationKeyframe getKeyframe(long time) {

		if (time == bufferKeyframeTime) {
			return bufferKeyframe;
		}

		if (time < keyframeTimes[0]) {
			bufferKeyframeTime = time;
			bufferKeyframe = null;
			return null;
		} else if (time >= keyframeTimes[keyframeTimes.length - 1]) {
			bufferKeyframeTime = time;
			bufferKeyframe = keyframes[keyframes.length - 1];
			return keyframes[keyframes.length - 1];
		} else {
			for (int i = 1; i < keyframeTimes.length; i++) {
				if (keyframeTimes[i] > time) {
					bufferKeyframeTime = time;
					bufferKeyframe = keyframes[i - 1];
					return keyframes[i - 1];
				}
			}

			bufferKeyframeTime = time;
			bufferKeyframe = null;
			return null;
		}
	}

	public void finalizeData() {
		long lastCompleted = 0;
		int traverseCount = 0;
		EdgeAnimationKeyframe currentFrame;
		List<TokenAnimation> currentAnims, lastAnims;
		lastAnims = new ArrayList<TokenAnimation>();
		for (int i = 1; i < keyframeList.size(); i++) {
			currentFrame = keyframeList.get(i);
			currentAnims = currentFrame.getTokenAnimations();
			for (TokenAnimation anim : lastAnims) {
				if (anim.getEnd() == currentFrame.getTime()) {
					// token animation has completed at this step
					traverseCount++;
					lastCompleted = currentFrame.getTime();
				}
			}
			currentFrame.setTraverseCount(traverseCount);
			currentFrame.setLastCompleted(lastCompleted);
			lastAnims = currentAnims;
		}
		// add first empty keyframe
		EdgeAnimationKeyframe firstFrame = new EdgeAnimationKeyframe(0);
		firstFrame.setLastCompleted(0);
		firstFrame.setTraverseCount(0);
		keyframeList.add(0, firstFrame);
		// create efficient data structures for lookup
		keyframeTimes = new long[keyframeList.size()];
		keyframes = new EdgeAnimationKeyframe[keyframeList.size()];
		for (int i = 0; i < keyframeList.size(); i++) {
			keyframes[i] = keyframeList.get(i);
			keyframeTimes[i] = keyframes[i].getTime();
		}
		// remove list (triggers null pointer exceptions on errors)
		keyframeList = null;
		// initialize buffer
		bufferKeyframeTime = -1;
		bufferKeyframe = null;
	}

	public int getTraverseCount() {
		return keyframes[keyframes.length - 1].getTraverseCount();
	}

	public void addTokenAnimation(TokenAnimation anim) {
		int startIndex = getKeyframeIndexAtTimeInserting(anim.getStart());
		int endIndex = getKeyframeIndexAtTimeInserting(anim.getEnd());
		for (int i = startIndex; i < endIndex; i++) {
			keyframeList.get(i).addTokenAnimation(anim);
		}
	}

	protected int getKeyframeIndexAtTimeInserting(long time) {
		int listSize = keyframeList.size();

		if (listSize == 0) {
			insertKeyframeAtPosition(0, time);
			return 0;
		}
		long lastTime = keyframeList.get(listSize - 1).getTime();
		long firstTime = keyframeList.get(0).getTime();
		if (time < firstTime) {
			insertKeyframeAtPosition(0, time);
			return 0;
		} else if (time > lastTime) {
			insertKeyframeAtPosition(listSize, time);
			return listSize;
		} else {
			long diffToLast = lastTime - time;
			long diffToFirst = time - firstTime;
			long currentTime;
			if (diffToFirst < diffToLast) {
				for (int i = 0; i < listSize; i++) {
					currentTime = keyframeList.get(i).getTime();
					if (currentTime == time) {
						return i;
					} else if (currentTime > time) {
						insertKeyframeAtPosition(i, time);
						return i;
					}
				}
			} else {
				for (int i = listSize - 1; i >= 0; i--) {
					currentTime = keyframeList.get(i).getTime();
					if (currentTime == time) {
						return i;
					} else if (currentTime < time) {
						insertKeyframeAtPosition(i + 1, time);
						return i + 1;
					}
				}
			}
		}
		throw new Error("unable to address correctly!");
	}

	protected void insertKeyframeAtPosition(int position, long time) {
		EdgeAnimationKeyframe keyFrame = new EdgeAnimationKeyframe(time);
		keyframeList.add(position, keyFrame);
		// copy relevant token animations from last frame
		if (position > 0) {
			List<TokenAnimation> tokenAnimations = keyframeList.get(position - 1).getTokenAnimations();
			for (TokenAnimation anim : tokenAnimations) {
				if (anim.getEnd() > time) {
					keyFrame.addTokenAnimation(anim);
				}
			}
		}
	}

}
