package org.processmining.models.animation;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

import org.processmining.models.graphbased.directed.AbstractDirectedGraphNode;

public class NodeAnimation {
	private final AbstractDirectedGraphNode node;
	private final Rectangle2D rect;

	private List<NodeAnimationKeyframe> keyframeList;

	private long[] keyframeTime;
	private NodeAnimationKeyframe[] keyframes;

	// buffers to speed up repeated lookups
	private long keyframeBufferTime;
	private NodeAnimationKeyframe keyframeBuffer;

	public NodeAnimation(AbstractDirectedGraphNode node, Rectangle2D rect) {
		this.node = node;
		this.rect = rect;
		keyframeList = new ArrayList<NodeAnimationKeyframe>();
	}

	public AbstractDirectedGraphNode getNode() {
		return node;
	}
	
	public double getPositionX() {
		return rect.getX();
	}

	public double getPositionY() {
		return rect.getY();
	}

	public double getWidth() {
		return rect.getWidth();
	}

	public double getHeight() {
		return rect.getHeight();
	}

	/**
	 * 
	 * @param start
	 *            The start time of the given interval (inclusive).
	 * @param end
	 *            The end time of the given interval (exclusive).
	 * @return The number of key frames for this state in the given time
	 *         interval.
	 */
	public int getActivityBetween(long start, long end) {
		int activity = 0;
		for (int i = 0; i < keyframeTime.length; i++) {
			if (keyframeTime[i] >= start) {
				if (keyframeTime[i] < end) {
					activity++;
				} else {
					break;
				}
			}
		}
		return activity;
	}

	/**
	 * Finalizes the key frames.
	 */
	public void finalizeData() {
		/*
		 * Insert a dummy first key frame.
		 */
		NodeAnimationKeyframe firstFrame = new NodeAnimationKeyframe(null, 0L, false, false);
		firstFrame.setEventCount(0);
		firstFrame.setSwallowCount(0);
		firstFrame.setEventCount(0);
		keyframeList.add(0, firstFrame);

		/*
		 * Fill keyframeTime. As a side effect, accumulate emerge (cases
		 * starting at this state) and swallow (cases terminating at this state)
		 * counts.
		 */
		keyframeTime = new long[keyframeList.size()];
		keyframes = new NodeAnimationKeyframe[keyframeList.size()];
		int emergeCount = 0;
		int swallowCount = 0;
		for (int i = 0; i < keyframeList.size(); i++) {
			keyframes[i] = keyframeList.get(i);
			keyframeTime[i] = keyframes[i].getTime();
			keyframes[i].setEventCount(i);
			keyframes[i].setEventCount(i);
			emergeCount += keyframes[i].getEmergeCount();
			keyframes[i].setEmergeCount(emergeCount);
			swallowCount += keyframes[i].getSwallowCount();
			keyframes[i].setSwallowCount(swallowCount);
		}
		keyframeList = null;

		/*
		 * Initialize buffer.
		 */
		keyframeBufferTime = -1;
		keyframeBuffer = null;
	}

	/**
	 * Inserts the given key frame at its proper place. Key frames should be
	 * sorted on ascending time.
	 * 
	 * @param keyframe
	 *            The given key frame.
	 */
	public void addKeyframe(NodeAnimationKeyframe keyframe) {
		for (int i = keyframeList.size(); i > 0; i--) {
			if (keyframe.getTime() >= keyframeList.get(i - 1).getTime()) {
				keyframeList.add(i, keyframe);
				return;
			}
		}
		// insert at zero position
		keyframeList.add(0, keyframe);
	}

	/**
	 * 
	 * @param time
	 *            The given time.
	 * @return The latest key frame that precedes the given time.
	 */
	public synchronized NodeAnimationKeyframe getKeyframe(long time) {

		if (time == keyframeBufferTime) {
			/*
			 * Use buffered result.
			 */
			return keyframeBuffer;
		}

		if (time < keyframeTime[0]) {
			/*
			 * No key frame precedes the given time. Return null.
			 */
			keyframeBufferTime = time;
			keyframeBuffer = null;
			return null;
		} else if (time >= keyframeTime[keyframeTime.length - 1]) {
			/*
			 * Every key frame precedes the given time. Return latest key frame.
			 */
			keyframeBufferTime = time;
			keyframeBuffer = keyframes[keyframes.length - 1];
			return keyframes[keyframes.length - 1];
		} else {
			for (int i = 1; i < keyframes.length; i++) {
				if (keyframeTime[i] > time) {
					/*
					 * Found first key frame following the given time. Return
					 * previous key frame.
					 */
					keyframeBufferTime = time;
					keyframeBuffer = keyframes[i - 1];
					return keyframes[i - 1];
				}
			}

			/*
			 * Not found. Return null.
			 */
			keyframeBufferTime = time;
			keyframeBuffer = null;
			return null;
		}
	}

}
