package org.processmining.models.animation.visualization;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;

import javax.swing.JComponent;

import org.processmining.models.animation.Animation;
import org.processmining.models.animation.visualization.painters.ColorSettings;
import org.processmining.models.animation.visualization.painters.EdgeBackgroundPainter;
import org.processmining.models.animation.visualization.painters.EdgeTrajectoryPainter;
import org.processmining.models.animation.visualization.painters.NodeBadgesPainter;
import org.processmining.models.animation.visualization.painters.NodePainter;
import org.processmining.models.animation.visualization.painters.TokenLabelPainter;
import org.processmining.models.animation.visualization.painters.TokenPainter;
import org.processmining.models.animation.visualization.painters.TokenRingsPainter;

public class AnimationCanvas extends JComponent implements AnimationListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = -7849178353091200368L;

	private static GraphicsConfiguration gc;

	protected double zoom = 0.5f;
	protected double width;
	protected double height;

	protected double leftBorder;
	protected double rightBorder;
	protected double topBorder;
	protected double bottomBorder;

	private double boxHeight;
	private double minLineWidth;
	private double maxLineWidth;
	private double borderWidth;
	private double fontSize;

	protected BufferedImage buffer;
	protected BufferedImage backBuffer;

	protected Animation animation;
	protected long animationTime;

	protected ColorSettings colorSettings;

	protected EdgeBackgroundPainter edgeBackgroundPainter;
	protected EdgeTrajectoryPainter edgeTrajectoryPainter;
	protected NodePainter nodePainter;
	protected TokenPainter tokenPainter;
	protected TokenLabelPainter tokenLabelPainter;
	protected NodeBadgesPainter nodeBadgesPainter;
	protected TokenRingsPainter tokenRingsPainter;

	public AnimationCanvas(Animation animation, ColorSettings settings) {
		this.animation = animation;
		this.colorSettings = settings;
		this.animationTime = animation.getStart();
		this.boxHeight = animation.getNodeAnimations().get(0)
				.getHeight();
		this.maxLineWidth = this.boxHeight * 0.56f;
		this.minLineWidth = this.boxHeight * 0.05f;
		this.borderWidth = this.boxHeight * 0.05f;
		this.fontSize = this.boxHeight * 0.23f;
		this.edgeBackgroundPainter = new EdgeBackgroundPainter(this);
		this.edgeTrajectoryPainter = new EdgeTrajectoryPainter(this);
		this.nodePainter = new NodePainter(this);
		this.tokenPainter = new TokenPainter(this);
		this.tokenLabelPainter = new TokenLabelPainter(this);
		this.nodeBadgesPainter = new NodeBadgesPainter(this);
		this.tokenRingsPainter = new TokenRingsPainter(this);
		setZoom(0.5f);
		this.addMouseListener(new MouseListener() {
			public void mouseClicked(MouseEvent e) {
				if (colorSettings.equals(ColorSettings.SETTINGS_NIGHT)) {
					colorSettings = ColorSettings.SETTINGS_DAY;
				} else {
					colorSettings = ColorSettings.SETTINGS_NIGHT;
				}
				paintBuffered();
			}

			public void mouseEntered(MouseEvent e) { /* ignore */
			}

			public void mouseExited(MouseEvent e) { /* ignore */
			}

			public void mousePressed(MouseEvent e) { /* ignore */
			}

			public void mouseReleased(MouseEvent e) { /* ignore */
			}
		});
	}

	public Animation getAnimation() {
		return this.animation;
	}

	public ColorSettings getColorSettings() {
		return this.colorSettings;
	}

	public void setZoom(float zoom) {
		this.zoom = zoom;
		leftBorder = animation.getLayout().getBoundsHeight() * zoom * 0.1;
		topBorder = leftBorder / 3;
		bottomBorder = topBorder;
		rightBorder = leftBorder * 3;
		this.width = animation.getLayout().getBoundsWidth() * zoom;
		this.height = animation.getLayout().getBoundsHeight() * zoom;
		int imgWidth = (int) (width + leftBorder + rightBorder);
		int imgHeight = (int) (height + topBorder + bottomBorder);
		Dimension size = new Dimension(imgWidth, imgHeight);
		this.setMinimumSize(size);
		this.setPreferredSize(size);
		this.buffer = createCompatibleImage(imgWidth, imgHeight);
		this.backBuffer = createCompatibleImage(imgWidth, imgHeight);
		revalidate();
		paintBuffered();
	}

	public double getBoxHeight() {
		return this.boxHeight;
	}

	public double getLineWidth(float percentage) {
		double lineWidth = (this.minLineWidth + (this.maxLineWidth * percentage))
				* zoom;

		if (lineWidth < 1.0f) {
			lineWidth = 1.0f;
		}
		return lineWidth;
	}

	public double getBorderWidth() {
		double bWidth = this.borderWidth;
		if (bWidth < 1.0f) {
			bWidth = 1.0f;
		}
		return bWidth;
	}

	public float getFontSize() {

		return (float) (this.fontSize * zoom);
	}

	public double translateX(double modelX) {
		// return modelX * this.width;
		return modelX * zoom;
	}

	public double translateY(double modelY) {
		// return modelY * this.height;
		return modelY * zoom;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.processmining.mining.fuzzymining.anim.model.AnimationListener#
	 * updateModelTime(long)
	 */
	public void updateModelTime(long modelTime) {
		this.animationTime = modelTime;
		paintBuffered();
	}

	protected void paintComponent(Graphics g) {
		int x = 0;
		int y = 0;
		int actualWidth = this.getWidth();
		int actualHeight = this.getHeight();
		int canvasWidth = (int) (this.width + leftBorder + rightBorder);
		int canvasHeight = (int) (this.height + topBorder + bottomBorder);
		if (actualWidth > canvasWidth) {
			x = (actualWidth - canvasWidth) / 2;
		}
		if (actualHeight > canvasHeight) {
			y = (actualHeight - canvasHeight) / 2;
		}
		g.setColor(colorSettings.getCanvasBackground());
		g.fillRect(0, 0, actualWidth + 1, actualHeight + 1);
		g.drawImage(buffer, x, y, null);
	}

	protected void paintBuffered() {
		long modelTime = this.animationTime;

		if (backBuffer != null) {
			Graphics2D g2d = backBuffer.createGraphics();
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);

			g2d.setColor(colorSettings.getCanvasBackground());
			g2d.fill(new Rectangle2D.Float(0, 0, backBuffer.getWidth() + 1,
					backBuffer.getHeight() + 1));

			g2d.translate(leftBorder, topBorder);

			Graphics2D g2dCopy = (Graphics2D) g2d.create();
			this.edgeBackgroundPainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.edgeTrajectoryPainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.nodePainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.tokenRingsPainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.tokenPainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.tokenLabelPainter.paint(modelTime, g2dCopy);

			g2dCopy.dispose();
			g2dCopy = (Graphics2D) g2d.create();
			this.nodeBadgesPainter.paint(modelTime, g2dCopy);

			BufferedImage tmp = this.backBuffer;
			this.backBuffer = this.buffer;
			this.buffer = tmp;

			if (this.isDisplayable()) {
				repaint();
			}
			g2d.dispose();
		}
	}

	protected BufferedImage createCompatibleImage(int width, int height) {
		if (gc == null) {
			gc = getGraphicsConfiguration();
			if (gc == null) {
				gc = GraphicsEnvironment.getLocalGraphicsEnvironment()
						.getDefaultScreenDevice().getDefaultConfiguration();
			}
		}
		return gc.createCompatibleImage(width, height);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mining.fuzzymining.vis.anim.AnimationListener#started()
	 */
	public void started() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.mining.fuzzymining.vis.anim.AnimationListener#stopped()
	 */
	public void stopped() {
	}
}
