package org.processmining.models.animation.visualization.painters;

import java.awt.Color;

public class ColorSettings {

	/**
	 * Mixes two colors.
	 * 
	 * @param color1 First color.
	 * @param color2 Second color.
	 * @param frac2 Fraction of second color to use.
	 * @return '(1-frac2)*color1 + frac2*color2'
	 */
	public static Color mix(Color color1, Color color2, float frac2) {
		if (frac2 <= 0f) {
			return color1;
		} else if (frac2 >= 1f) {
			return color2;
		} else {
			float frac1 = 1f - frac2;
			float red = (frac1 * color1.getRed()) + (frac2 * color2.getRed());
			float green = (frac1 * color1.getGreen()) + (frac2 * color2.getGreen());
			float blue = (frac1 * color1.getBlue()) + (frac2 * color2.getBlue());
			float alpha = (frac1 * color1.getAlpha()) + (frac2 * color2.getAlpha());
			return new Color(red / 255f, green / 255f, blue / 255f, alpha / 255f);
		}
	}

	/**
	 * Fades a color. Not equivalent to mixing the color with black, as the 
	 * alpha component is not effected by fading. 
	 * 
	 * @param color Color to fade.
	 * @param frac Fraction of color to fade to.
	 * @return 'frac*color'
	 */
	public static Color fade(Color color, float frac) {
		float red = color.getRed() * frac;
		float green = color.getGreen() * frac;
		float blue = color.getBlue() * frac;
		return new Color(red / 255f, green / 255f, blue / 255f, color.getAlpha() / 255f);
	}

	/**
	 * Mixes the color with a darker shade of grey.
	 * 
	 * @param color Color to mix
	 * @param frac Fraction of color to retain (1 = return color, 0 = return darker shade of grey
	 * @return
	 */
	public static Color darken(Color color, float frac) {
		float dkGreyFrac = 1f - frac;
		/*
		 * Get corresponding grey.
		 */
		int dkGrey = (color.getRed() + color.getGreen() + color.getBlue()) / 3;
		/*
		 * Get darker shade of grey.
		 */
		dkGrey -= 30;
		if (dkGrey <= 0) {
			dkGrey = 0;
		}
		/*
		 * Mix color with darker shade of grey.
		 */
		float dkGreyVal = dkGrey * dkGreyFrac;
		float red = (color.getRed() * frac) + dkGreyVal;
		float green = (color.getGreen() * frac) + dkGreyVal;
		float blue = (color.getBlue() * frac) + dkGreyVal;
		return new Color(red /255f, green / 255f, blue / 255f);
	}

	/**
	 * Sets the alpha component of the color.
	 * 
	 * @param color
	 * @param alpha
	 * @return The given color (RGB), with the given alpha component.
	 */
	public static Color setAlpha(Color color, float alpha) {
		return new Color(color.getRed(), color.getGreen(), color.getBlue(), (int) (alpha * 255));
	}

	/*
	 * Color schemes to use.
	 */
	protected Color canvasBackground = new Color(0, 0, 0);

	protected Color passiveEdgeBackground = new Color(40, 40, 40);
	protected Color activeEdgeBackground = new Color(200, 0, 0);
	protected Color activeEdgeTrajectoryStart = new Color(200, 0, 0, 20);

	protected Color animRingsBackground = new Color(200, 200, 200);

	protected Color tokensBackground = new Color(200, 200, 200);

	protected Color passiveStateBackground = new Color(30, 30, 30);
	protected Color passiveStateForeground = new Color(100, 100, 100);
	protected Color passiveStateBorder = new Color(80, 80, 80);
	protected Color activeStateBackground = new Color(190, 0, 0);
	protected Color activeStateForeground = new Color(220, 220, 220);
	protected Color activeStateBorder = new Color(180, 180, 180);

	protected Color tokenLabel = new Color(220, 180, 0, 160);

	protected Color nodeBadgeForeground = new Color(220, 220, 220, 220);
	protected Color nodeBadgeBackgroundSwallowed = new Color(200, 0, 0, 200);
	protected Color nodeBadgeBackgroundEmerged = new Color(0, 180, 0, 200);
	protected Color nodeBadgeBackgroundTraversed = new Color(0, 0, 200, 200);

	protected Color tokenRing = new Color(200, 200, 200);

	public ColorSettings() {
		// standard constructor
	}

	/*
	 * Default color scheme.
	 */
	public static ColorSettings SETTINGS_NIGHT = new ColorSettings();

	/*
	 * Alternative colors scheme.
	 */
	public static ColorSettings SETTINGS_DAY = new ColorSettings(new Color(255, 255, 255), new Color(120, 120, 120),
			new Color(255, 30, 0), new Color(255, 30, 30), new Color(0, 0, 0), new Color(210, 210, 210), new Color(80,
					80, 80), new Color(80, 80, 80), new Color(200, 20, 0), new Color(0, 0, 0), new Color(0, 0, 0),
			new Color(80, 60, 40, 200), new Color(20, 20, 20, 220), new Color(200, 0, 0, 200),
			new Color(0, 180, 0, 200), new Color(0, 0, 200, 200), new Color(255, 0, 0, 20), new Color(180, 40, 0));

	/*
	 * Sets specific color scheme.
	 */
	public ColorSettings(Color canvasBackground, Color passiveEdgeBackground, Color activeEdgeBackground,
			Color animRingsBackground, Color tokensBackground, Color passiveStateBackground,
			Color passiveStateForeground, Color passiveStateBorder, Color activeStateBackground,
			Color activeStateForeground, Color activeStateBorder, Color tokenLabel, Color nodeBadgeForeground,
			Color nodeBadgeBackgroundSwallowed, Color nodeBadgeBackgroundEmerged, Color nodeBadgeBackgroundTraversed,
			Color activeEdgeTrajectoryStart, Color tokenRing) {
		super();
		this.canvasBackground = canvasBackground;
		this.passiveEdgeBackground = passiveEdgeBackground;
		this.activeEdgeBackground = activeEdgeBackground;
		this.animRingsBackground = animRingsBackground;
		this.tokensBackground = tokensBackground;
		this.passiveStateBackground = passiveStateBackground;
		this.passiveStateForeground = passiveStateForeground;
		this.passiveStateBorder = passiveStateBorder;
		this.activeStateBackground = activeStateBackground;
		this.activeStateForeground = activeStateForeground;
		this.activeStateBorder = activeStateBorder;
		this.tokenLabel = tokenLabel;
		this.nodeBadgeForeground = nodeBadgeForeground;
		this.nodeBadgeBackgroundSwallowed = nodeBadgeBackgroundSwallowed;
		this.nodeBadgeBackgroundEmerged = nodeBadgeBackgroundEmerged;
		this.nodeBadgeBackgroundTraversed = nodeBadgeBackgroundTraversed;
		this.activeEdgeTrajectoryStart = activeEdgeTrajectoryStart;
		this.tokenRing = tokenRing;
	}

	public Color getCanvasBackground() {
		return canvasBackground;
	}

	/**
	 * @return the passiveTransitionBackground
	 */
	public Color getPassiveEdgeBackground() {
		return passiveEdgeBackground;
	}

	/**
	 * @return the activeTransitionBackground
	 */
	public Color getActiveEdgeBackground() {
		return activeEdgeBackground;
	}

	public Color getActiveEdgeTrajectoryStart() {
		return activeEdgeTrajectoryStart;
	}

	/**
	 * @return the animRingsBackground
	 */
	public Color getAnimRingsBackground() {
		return animRingsBackground;
	}

	/**
	 * @return the tokensBackground
	 */
	public Color getTokensBackground() {
		return tokensBackground;
	}

	/**
	 * @return the passiveStateBackground
	 */
	public Color getPassiveStateBackground() {
		return passiveStateBackground;
	}

	/**
	 * @return the passiveStateForeground
	 */
	public Color getPassiveStateForeground() {
		return passiveStateForeground;
	}

	public Color getPassiveStateBorder() {
		return passiveStateBorder;
	}

	/**
	 * @return the activeStateBackground
	 */
	public Color getActiveStateBackground() {
		return activeStateBackground;
	}

	/**
	 * @return the activeStateForeground
	 */
	public Color getActiveStateForeground() {
		return activeStateForeground;
	}

	public Color getActiveStateBorder() {
		return activeStateBorder;
	}

	/**
	 * @return the tokenLabel
	 */
	public Color getTokenLabel() {
		return tokenLabel;
	}

	/**
	 * @return the taskBadgeForeground
	 */
	public Color getNodeBadgeForeground() {
		return nodeBadgeForeground;
	}

	/**
	 * @return the taskBadgeBackgroundSwallowed
	 */
	public Color getNodeBadgeBackgroundSwallowed() {
		return nodeBadgeBackgroundSwallowed;
	}

	/**
	 * @return the taskBackgroundEmerged
	 */
	public Color getNodeBadgeBackgroundEmerged() {
		return nodeBadgeBackgroundEmerged;
	}

	/**
	 * @return the taskBackgroundTraversed
	 */
	public Color getNodeBadgeBackgroundTraversed() {
		return nodeBadgeBackgroundTraversed;
	}

	public Color getTokenRing() {
		return tokenRing;
	}
}
