package org.processmining.models.animation.visualization.painters;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.util.ArrayList;
import java.util.List;

import org.processmining.models.animation.NodeAnimation;
import org.processmining.models.animation.NodeAnimationKeyframe;
import org.processmining.models.animation.visualization.AnimationCanvas;

public class NodeBadgesPainter extends Painter {

	/**
	 * @param canvas
	 */
	public NodeBadgesPainter(AnimationCanvas canvas) {
		super(canvas);
	}

	/* (non-Javadoc)
	 * @see org.processmining.mining.fuzzymining.vis.paint.Painter#paint(long, java.awt.Graphics2D)
	 */
	@Override
	public void paint(long modelTime, Graphics2D g2d) {
		float fontSize = canvas.getFontSize();
		g2d.setFont(g2d.getFont().deriveFont(fontSize));
		FontMetrics fMetrics = g2d.getFontMetrics();
		long fadeTime = canvas.getAnimation().getFadeTime() * 2;
		List<NodeAnimation> taskAnimations = new ArrayList<NodeAnimation>();
		taskAnimations.addAll(canvas.getAnimation().getNodeAnimations());
		NodeAnimationKeyframe keyframe = null;
		for(NodeAnimation taskAnimation : taskAnimations) {
			keyframe = taskAnimation.getKeyframe(modelTime);
			long idleTime = modelTime - keyframe.getTime();
			Color emergedBgColor, swallowedBgColor, traversedBgColor, textColor;
			float percentage;
			if(idleTime >= fadeTime) {
				percentage = 0.7f;
			} else {
				percentage = (float)idleTime / (float)fadeTime;
				percentage = 1f - (percentage * 0.3f);
			}
			// derive colors
			emergedBgColor = ColorSettings.fade(canvas.getColorSettings().getNodeBadgeBackgroundEmerged(), percentage);
			swallowedBgColor = ColorSettings.fade(canvas.getColorSettings().getNodeBadgeBackgroundSwallowed(), percentage);
			traversedBgColor = ColorSettings.fade(canvas.getColorSettings().getNodeBadgeBackgroundTraversed(), percentage);
			textColor = ColorSettings.fade(canvas.getColorSettings().getNodeBadgeForeground(), percentage);
			// derive node coords
			double x = canvas.translateX(taskAnimation.getPositionX());
			double y = canvas.translateY(taskAnimation.getPositionY());
			double width = canvas.translateX(taskAnimation.getWidth());
			double height = canvas.translateY(taskAnimation.getHeight());
			double badgeHeight = height / 3;
			double badgeHBorder = badgeHeight / 3;
			double badgeRadius = badgeHeight / 3;
			double overlap = height / 4;
			// get counter values
			int emergeCount = keyframe.getEmergeCount();
			int swallowCount = keyframe.getSwallowCount();
			int traverseCount = keyframe.getEventCount();
			if(emergeCount > 0) {
				// paint emerge badge on upper right
				String emerged = Integer.toString(keyframe.getEmergeCount());
				Rectangle2D emergedBounds = fMetrics.getStringBounds(emerged, g2d);
				g2d.setColor(emergedBgColor);
				double badgeX = x + width - overlap;
				double badgeY = y + badgeHBorder;
				RoundRectangle2D.Double rrect = new RoundRectangle2D.Double(
						badgeX, 
						badgeY, 
						(int)emergedBounds.getWidth() + badgeHBorder + badgeHBorder, 
						badgeHeight, 
						badgeRadius, badgeRadius);
				g2d.fill(rrect);
				// triangle pointer
				GeneralPath triangle = new GeneralPath();
				triangle.moveTo((float)(badgeX + 0.5f), (float)(badgeY + badgeHBorder));
				triangle.lineTo((float)(badgeX - badgeHBorder), (float)(badgeY + (badgeHeight / 2)));
				triangle.lineTo((float)(badgeX + 0.5f), (float)(badgeY + badgeHeight - badgeHBorder));
				triangle.closePath();
				g2d.fill(triangle);
				// paint string
				double sX = badgeX + badgeHBorder;
				double sY = badgeY + (badgeHeight / 2) + (2 * (emergedBounds.getHeight() / 5));
				g2d.setColor(textColor);
				g2d.drawString(emerged, (int)sX, (int)sY);
			}
			if(swallowCount > 0) {
				// paint swallow badge on lower right
				String swallowed = Integer.toString(swallowCount);
				Rectangle2D swallowedBounds = fMetrics.getStringBounds(swallowed, g2d);
				g2d.setColor(swallowedBgColor);
				double badgeX = x + width - overlap;
				double badgeY = y + height - badgeHeight - badgeHBorder;
				RoundRectangle2D.Double rrect = new RoundRectangle2D.Double(
						badgeX, 
						badgeY, 
						(int)swallowedBounds.getWidth() + badgeHBorder + badgeHBorder, 
						badgeHeight, 
						badgeRadius, badgeRadius);
				g2d.fill(rrect);
				// triangle pointer
				GeneralPath triangle = new GeneralPath();
				triangle.moveTo((float)(badgeX + 0.5f), (float)(badgeY + badgeHBorder));
				triangle.lineTo((float)(badgeX - badgeHBorder), (float)(badgeY + (badgeHeight / 2)));
				triangle.lineTo((float)(badgeX + 0.5f), (float)(badgeY + badgeHeight - badgeHBorder));
				triangle.closePath();
				g2d.fill(triangle);
				// paint string
				double sX = badgeX + badgeHBorder;
				double sY = badgeY + (badgeHeight / 2) + (2 * (swallowedBounds.getHeight() / 5));
				g2d.setColor(textColor);
				g2d.drawString(swallowed, (int)sX, (int)sY);
			}
			if(traverseCount > 0) {
				// paint traverse badge on middle left
				String traversed = Integer.toString(traverseCount);
				Rectangle2D traversedBounds = fMetrics.getStringBounds(traversed, g2d);
				g2d.setColor(traversedBgColor);
				double badgeWidth = traversedBounds.getWidth() + badgeHBorder + badgeHBorder;
				double badgeX = x - badgeWidth + overlap;
				double badgeY = y + (height / 2) - (badgeHeight / 2);
				RoundRectangle2D.Double rrect = new RoundRectangle2D.Double(
						badgeX, 
						badgeY, 
						badgeWidth, 
						badgeHeight, 
						badgeRadius, badgeRadius);
				g2d.fill(rrect);
				// triangle pointer
				GeneralPath triangle = new GeneralPath();
				triangle.moveTo((float)(badgeX + badgeWidth), (float)(badgeY + badgeHBorder));
				triangle.lineTo((float)(badgeX + badgeWidth + badgeHBorder), (float)(badgeY + (badgeHeight / 2)));
				triangle.lineTo((float)(badgeX + badgeWidth) , (float)(badgeY + badgeHeight - badgeHBorder));
				triangle.closePath();
				g2d.fill(triangle);
				// paint string
				double sX = badgeX + badgeHBorder;
				double sY = badgeY + (badgeHeight / 2) + (2 * (traversedBounds.getHeight() / 5));
				g2d.setColor(textColor);
				g2d.drawString(traversed, (int)sX, (int)sY);
			}
		}
	}

}
